from enum import Enum, auto
from typing import Dict, Any, Optional, Literal
from typing import Union, List

from aiogram.dispatcher.fsm.state import State
from aiogram.types import (
    Message, User, CallbackQuery, Chat, ChatMemberUpdated, TelegramObject,
    Update,
)

DIALOG_EVENT_NAME = "aiogd_update"
Data = Union[Dict, List, int, str, None]


class StartMode(Enum):
    NORMAL = auto()
    RESET_STACK = auto()
    NEW_STACK = auto()


class Action(Enum):
    DONE = "DONE"
    START = "START"
    UPDATE = "UPDATE"
    SWITCH = "SWITCH"


class DialogUpdateEvent(TelegramObject):
    class Config:
        arbitrary_types_allowed = True
        use_enum_values = False

    from_user: User
    chat: Chat
    action: Action
    data: Any
    intent_id: Optional[str]
    stack_id: Optional[str]


class DialogStartEvent(DialogUpdateEvent):
    new_state: State
    mode: StartMode


class DialogSwitchEvent(DialogUpdateEvent):
    new_state: State


ChatEvent = Union[CallbackQuery, Message, DialogUpdateEvent, ChatMemberUpdated]


class DialogUpdate(Update):
    aiogd_update: DialogUpdateEvent

    def __init__(self, aiogd_update: DialogUpdateEvent):
        super().__init__(update_id=0, aiogd_update=aiogd_update)

    @property
    def event_type(self) -> str:
        return DIALOG_EVENT_NAME

    @property
    def event(self) -> DialogUpdateEvent:
        return self.aiogd_update


class FakeUser(User):
    fake: Literal[True] = True


class FakeChat(Chat):
    fake: Literal[True] = True



from .base import Keyboard
from .button import Button, Url
from .calendar_kbd import Calendar, ManagedCalendarAdapter
from .checkbox import Checkbox, ManagedCheckboxAdapter
from .group import Group, Row, Column
from .scrolling_group import ScrollingGroup, ManagedScrollingGroupAdapter
from .select import (
    Select, Radio, Multiselect, ManagedMultiSelectAdapter, ManagedRadioAdapter,
)
from .state import Back, Cancel, Next, Start, SwitchTo

__all__ = [
    "Keyboard",
    "Button", "Url",
    "Calendar", "ManagedCalendarAdapter",
    "Back", "Cancel", "Next", "Start", "SwitchTo",
    "Group", "Row", "Column",
    "ScrollingGroup", "ManagedScrollingGroupAdapter",
    "Checkbox", "ManagedCheckboxAdapter",
    "Select", "Radio", "Multiselect",
    "ManagedMultiSelectAdapter", "ManagedRadioAdapter",
]



import warnings
from typing import Optional, Any


def manager_deprecated(manager: Optional[Any]):
    if manager is not None:
        warnings.warn(
            "Passing `DialogManager` instance directly is deprecated"
            " and will be removed in aiogram_dialog==2.0",
            DeprecationWarning,
            stacklevel=3,
        )



import html
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

from aiogram.dispatcher.filters.state import State, StatesGroup
from aiogram.types import User, Chat, Message, ContentType
from jinja2 import Environment, PackageLoader, select_autoescape

from aiogram_dialog import DialogRegistry, DialogManager, Dialog
from aiogram_dialog.context.context import Context
from aiogram_dialog.context.events import DialogUpdateEvent, Action
from aiogram_dialog.context.stack import Stack
from aiogram_dialog.manager.protocols import (
    NewMessage, DialogRegistryProto, MediaAttachment,
)


@dataclass
class RenderButton:
    state: str
    title: str


@dataclass
class RenderWindow:
    message: str
    state: str
    keyboard: List[List[RenderButton]]
    photo: Optional[str]


@dataclass
class RenderDialog:
    state_group: str
    windows: List[RenderWindow]


class FakeManager(DialogManager):
    def __init__(self, registry: DialogRegistry):
        self.event = DialogUpdateEvent(
            from_user=User(id=1, is_bot=False, first_name="Fake"),
            chat=Chat(id=1, type="private"),
            action=Action.UPDATE,
            data={},
            intent_id=None,
            stack_id=None,
        )
        self._registry = registry
        self._context: Optional[Context] = None
        self._dialog = None
        self.data = {
            "dialog_manager": self,
            "event_chat": Chat(id=1, type="private"),
            "event_from_user": User(id=1, is_bot=False, first_name="Fake")
        }

    def set_dialog(self, dialog: Dialog):
        self._dialog = dialog
        self.reset_context()

    def set_state(self, state: State):
        self._context.state = state

    def dialog(self) -> Dialog:
        return self._dialog

    def is_preview(self) -> bool:
        return True

    def reset_context(self) -> None:
        self._context = Context(
            _intent_id="0",
            _stack_id="0",
            start_data={},
            widget_data={},
            dialog_data={},
            state=State(),
        )

    def current_stack(self) -> Optional[Stack]:
        return Stack()

    def current_context(self) -> Optional[Context]:
        return self._context

    async def show(self, new_message: NewMessage) -> Message:
        self.new_message = new_message
        return Message(message_id=1, date=datetime.now(), chat=Chat(id=1, type="private"))

    @property
    def registry(self) -> DialogRegistryProto:
        return self._registry


def create_photo(media: Optional[MediaAttachment]) -> Optional[str]:
    if not media:
        return
    if media.type != ContentType.PHOTO:
        return
    if media.url:
        return media.url
    if media.path:
        return media.path
    if media.file_id:
        return str(media.file_id)


def create_window(state: State, msg: NewMessage) -> RenderWindow:
    if msg.parse_mode is None or msg.parse_mode == "None":
        text = html.escape(msg.text)
    else:
        text = msg.text

    return RenderWindow(
        message=text.replace("\n", "<br>"),
        state=state.state,
        photo=create_photo(media=msg.media),
        keyboard=[
            [
                RenderButton(title=button.text, state=button.callback_data)
                for button in row
            ]
            for row in msg.reply_markup.inline_keyboard
        ]
    )


async def render_dialog(manager: FakeManager, group: StatesGroup,
                        dialog: Dialog) -> RenderDialog:
    manager.set_dialog(dialog)
    windows = []
    for state in group.__states__:
        manager.set_state(state)
        await dialog.show(manager)
        windows.append(create_window(state, manager.new_message))

    return RenderDialog(state_group=str(group), windows=windows)


async def render_preview(registry: DialogRegistry, file: str):
    fake_manager = FakeManager(registry)
    dialogs = [
        await render_dialog(fake_manager, group, dialog)
        for group, dialog in registry.dialogs.items()
    ]
    env = Environment(
        loader=PackageLoader("aiogram_dialog.tools"),
        autoescape=select_autoescape()
    )
    template = env.get_template("message.html")
    res = template.render(dialogs=dialogs)
    with open(file, "w") as f:
        f.write(res)



from logging import getLogger
from typing import Optional, Tuple, Union

from aiogram import Bot
from aiogram.exceptions import TelegramBadRequest
from aiogram.types import (
    FSInputFile, InputFile, URLInputFile,
    Message, CallbackQuery, Chat, ChatMemberUpdated, ContentType, InputMedia,
    User,
)

from .context.events import (
    DialogUpdateEvent, ChatEvent
)
from .manager.protocols import MediaAttachment, NewMessage, ShowMode, MediaId

logger = getLogger(__name__)

CB_SEP = "\x1D"

send_methods = {
    ContentType.ANIMATION: "send_animation",
    ContentType.AUDIO: "send_audio",
    ContentType.DOCUMENT: "send_document",
    ContentType.PHOTO: "send_photo",
    ContentType.VIDEO: "send_video",
}


def get_chat(event: ChatEvent) -> Chat:
    if isinstance(event, (Message, DialogUpdateEvent, ChatMemberUpdated)):
        return event.chat
    elif isinstance(event, CallbackQuery):
        if not event.message:
            return Chat(id=event.from_user.id, type='')
        return event.message.chat


def is_chat_loaded(chat: Chat) -> bool:
    """
    Checks if chat is correctly loaded from telegram.
    Otherwise, it is created with no data inside as a FakeChat
    """
    return getattr(chat, "fake", False)


def is_user_loaded(user: User) -> bool:
    """
    Checks if chat is correctly loaded from telegram.
    Otherwise, it is created with no data inside as a FakeUser
    """
    return getattr(user, "fake", False)


def get_media_id(message: Message) -> Optional[MediaId]:
    media = (
            message.audio or
            message.animation or
            message.document or
            (message.photo[-1] if message.photo else None) or
            message.video
    )
    if not media:
        return None
    return MediaId(
        file_id=media.file_id,
        file_unique_id=media.file_unique_id,
    )


async def get_media_source(media: MediaAttachment) -> Union[InputFile, str]:
    if media.file_id:
        return media.file_id.file_id
    if media.url:
        return URLInputFile(media.url)
    else:
        return FSInputFile(media.path)


def intent_callback_data(intent_id: str,
                         callback_data: Optional[str]) -> Optional[str]:
    if callback_data is None:
        return None
    return intent_id + CB_SEP + callback_data


def add_indent_id(message: NewMessage, intent_id: str):
    if not message.reply_markup:
        return
    for row in message.reply_markup.inline_keyboard:
        for button in row:
            button.callback_data = intent_callback_data(
                intent_id, button.callback_data
            )


def remove_indent_id(callback_data: str) -> Tuple[str, str]:
    if CB_SEP in callback_data:
        intent_id, new_data = callback_data.split(CB_SEP, maxsplit=1)
        return intent_id, new_data
    return "", callback_data


async def show_message(bot: Bot, new_message: NewMessage,
                       old_message: Optional[Message]):
    if not old_message or new_message.show_mode is ShowMode.SEND:
        logger.debug("Send new message, because: mode=%s, has old_message=%s",
                     new_message.show_mode, bool(old_message))
        await remove_kbd(bot, old_message)
        return await send_message(bot, new_message)

    had_media = old_message.content_type != ContentType.TEXT
    need_media = bool(new_message.media)

    if (
            new_message.text == old_message.text and
            new_message.reply_markup == old_message.reply_markup and
            had_media == need_media and
            (
                    not need_media or
                    new_message.media.file_id == get_media_id(old_message)
            )
    ):
        # nothing changed: text, keyboard or media
        return old_message

    if had_media != need_media:
        # we cannot edit message if media appeared or removed
        try:
            await bot.delete_message(chat_id=old_message.chat.id,
                                     message_id=old_message.message_id)
        except TelegramBadRequest as err:
            if (
                    'message to delete not found' in err.message or
                    'message can\'t be deleted' in err.message
            ):
                await remove_kbd(bot, old_message)
            else:
                raise
        return await send_message(bot, new_message)

    try:
        return await edit_message(bot, new_message, old_message)
    except TelegramBadRequest as err:
        if 'message is not modified' in err.message:
            return old_message
        if (
                'message can\'t be edited' in err.message or
                'message to edit not found' in err.message
        ):
            return await send_message(bot, new_message)
        else:
            raise


async def remove_kbd(bot: Bot, old_message: Optional[Message]):
    if old_message:
        logger.debug("remove_kbd in %s", old_message.chat)
        try:
            await bot.edit_message_reply_markup(
                message_id=old_message.message_id,
                chat_id=old_message.chat.id,
            )
        except TelegramBadRequest as err:
            if 'message is not modified' in err.message:
                pass
            elif 'message can\'t be edited' in err.message:
                pass
            elif 'message to edit not found' in err.message:
                pass
            else:
                raise err


# Edit
async def edit_message(bot: Bot, new_message: NewMessage,
                       old_message: Message):
    if new_message.media:
        if new_message.media.file_id == get_media_id(old_message):
            return await edit_caption(bot, new_message, old_message)
        return await edit_media(bot, new_message, old_message)
    else:
        return await edit_text(bot, new_message, old_message)


async def edit_caption(bot: Bot, new_message: NewMessage,
                       old_message: Message):
    logger.debug("edit_caption to %s", new_message.chat)
    return await bot.edit_message_caption(
        message_id=old_message.message_id,
        chat_id=old_message.chat.id,
        caption=new_message.text,
        reply_markup=new_message.reply_markup,
        parse_mode=new_message.parse_mode,
    )


async def edit_text(bot: Bot, new_message: NewMessage,
                    old_message: Message):
    logger.debug("edit_text to %s", new_message.chat)
    return await bot.edit_message_text(
        message_id=old_message.message_id,
        chat_id=old_message.chat.id,
        text=new_message.text,
        reply_markup=new_message.reply_markup,
        parse_mode=new_message.parse_mode,
        disable_web_page_preview=new_message.disable_web_page_preview,
    )


async def edit_media(bot: Bot, new_message: NewMessage,
                     old_message: Message):
    logger.debug("edit_media to %s, media_id: %s",
                 new_message.chat, new_message.media.file_id)
    media = InputMedia(
        caption=new_message.text,
        reply_markup=new_message.reply_markup,
        parse_mode=new_message.parse_mode,
        disable_web_page_preview=new_message.disable_web_page_preview,
        media=await get_media_source(new_message.media),
        **new_message.media.kwargs,
    )
    return await bot.edit_message_media(
        message_id=old_message.message_id,
        chat_id=old_message.chat.id,
        media=media,
    )


# Send
async def send_message(bot: Bot, new_message: NewMessage):
    if new_message.media:
        return await send_media(bot, new_message)
    else:
        return await send_text(bot, new_message)


async def send_text(bot: Bot, new_message: NewMessage):
    logger.debug("send_text to %s", new_message.chat)
    return await bot.send_message(
        new_message.chat.id,
        text=new_message.text,
        disable_web_page_preview=new_message.disable_web_page_preview,
        reply_markup=new_message.reply_markup,
        parse_mode=new_message.parse_mode,
    )


async def send_media(bot: Bot, new_message: NewMessage):
    logger.debug("send_media to %s, media_id: %s",
                 new_message.chat, new_message.media.file_id)
    method = getattr(bot, send_methods[new_message.media.type], None)
    if not method:
        raise ValueError(
            f"ContentType {new_message.media.type} is not supported")
    return await method(
        new_message.chat.id,
        await get_media_source(new_message.media),
        caption=new_message.text,
        reply_markup=new_message.reply_markup,
        parse_mode=new_message.parse_mode,
        **new_message.media.kwargs,
    )



from typing import Any, Callable, Dict, Awaitable, Union, cast

from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.types import Update

from .manager import ManagerImpl
from .protocols import DialogManager, DialogRegistryProto
from ..context.events import DialogUpdateEvent, ChatEvent

MANAGER_KEY = "dialog_manager"


class ManagerMiddleware(BaseMiddleware):
    def __init__(self, registry: DialogRegistryProto):
        super().__init__()
        self.registry = registry

    async def __call__(
            self,
            handler: Callable[[Union[Update, DialogUpdateEvent], Dict[str, Any]], Awaitable[Any]],
            event: ChatEvent,
            data: Dict[str, Any],
    ) -> Any:
        data[MANAGER_KEY] = ManagerImpl(
            event=event,
            registry=self.registry,
            data=data,
        )

        try:
            return await handler(event, data)
        finally:
            manager: DialogManager = data.pop(MANAGER_KEY, None)
            if manager:
                await manager.close_manager()



from logging import getLogger
from typing import Any, Dict, Optional

from aiogram import Bot
from aiogram.dispatcher.fsm.state import State
from aiogram.types import Chat, User

from .protocols import DialogRegistryProto, BaseDialogManager
from ..context.events import (
    Data, Action, DialogStartEvent, DialogSwitchEvent, DialogUpdateEvent,
    StartMode, DialogUpdate,  FakeChat, FakeUser,
)
from ..context.stack import DEFAULT_STACK_ID
from ..utils import is_chat_loaded, is_user_loaded

logger = getLogger(__name__)


class BgManager(BaseDialogManager):
    def __init__(
            self,
            user: User,
            chat: Chat,
            bot: Bot,
            registry: DialogRegistryProto,
            intent_id: Optional[str],
            stack_id: Optional[str],
            load: bool = False,
    ):
        self.user = user
        self.chat = chat
        self.bot = bot
        self._registry = registry
        self.intent_id = intent_id
        self.stack_id = stack_id
        self.load = load

    @property
    def registry(self) -> DialogRegistryProto:
        return self._registry

    def bg(
            self,
            user_id: Optional[int] = None,
            chat_id: Optional[int] = None,
            stack_id: Optional[str] = None,
            load: bool = False,
    ) -> "BaseDialogManager":
        if chat_id in (None, self.chat.id):
            chat = self.chat
        else:
            chat = FakeChat(id=chat_id, type="")

        if user_id in (None, self.user.id):
            user = self.user
        else:
            user = FakeUser(id=user_id, is_bot=False, first_name="")

        same_chat = (user.id == self.user.id and chat.id == self.chat.id)
        if stack_id is None:
            if same_chat:
                stack_id = self.stack_id
                intent_id = self.intent_id
            else:
                stack_id = DEFAULT_STACK_ID
                intent_id = None
        else:
            intent_id = None

        return BgManager(
            user=user,
            chat=chat,
            bot=self.bot,
            registry=self.registry,
            intent_id=intent_id,
            stack_id=stack_id,
            load=load,
        )

    def _base_event_params(self):
        return {
            "from_user": self.user,
            "chat": self.chat,
            "intent_id": self.intent_id,
            "stack_id": self.stack_id,
        }

    async def _notify(self, event: DialogUpdateEvent):
        await self.registry.notify(
            bot=self.bot,
            update=DialogUpdate(
                aiogd_update=event
            )
        )

    async def _load(self):
        if self.load:
            if not is_chat_loaded(self.chat):
                logger.debug("load chat: %s", self.chat.id)
                self.chat = await self.bot.get_chat(self.chat.id)
            if not is_user_loaded(self.user):
                logger.debug(
                    "load user %s from chat %s",
                    self.chat.id,
                    self.user.id
                )
                chat_member = await self.bot.get_chat_member(
                    self.chat.id,
                    self.user.id
                )
                self.user = chat_member.user

    async def done(self, result: Any = None) -> None:
        await self._load()
        await self._notify(DialogUpdateEvent(
            action=Action.DONE,
            data=result,
            **self._base_event_params()
        ))

    async def start(self, state: State, data: Data = None,
                    mode: StartMode = StartMode.NORMAL) -> None:
        await self._load()
        await self._notify(DialogStartEvent(
            action=Action.START,
            data=data,
            new_state=state,
            mode=mode,
            **self._base_event_params()
        ))

    async def switch_to(self, state: State) -> None:
        await self._load()
        await self._notify(DialogSwitchEvent(
            action=Action.SWITCH,
            data={},
            new_state=state,
            **self._base_event_params()
        ))

    async def update(self, data: Dict) -> None:
        await self._load()
        await self._notify(DialogUpdateEvent(
            action=Action.UPDATE,
            data=data,
            **self._base_event_params()
        ))



from logging import getLogger
from typing import Optional, Type, Dict, Union, Any, Callable, Awaitable

from aiogram.dispatcher.event.bases import CancelHandler
from aiogram.dispatcher.filters.base import BaseFilter
from aiogram.dispatcher.fsm.state import StatesGroup
from aiogram.dispatcher.fsm.storage.base import BaseStorage
from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.types import (
    Message, CallbackQuery, Update, TelegramObject, ChatMemberUpdated,
)

from .context import Context
from .events import DialogUpdateEvent, DialogUpdate
from .storage import StorageProxy
from ..exceptions import InvalidStackIdError, OutdatedIntent
from ..utils import remove_indent_id

STORAGE_KEY = "aiogd_storage_proxy"
STACK_KEY = "aiogd_stack"
CONTEXT_KEY = "aiogd_context"
CALLBACK_DATA_KEY = "aiogd_original_callback_data"

logger = getLogger(__name__)


class IntentFilter(BaseFilter):
    aiogd_intent_state_group: Optional[Type[StatesGroup]]

    async def __call__(
            self, obj: TelegramObject, **kwargs
    ) -> bool:
        if self.aiogd_intent_state_group is None:
            return True

        context: Context = kwargs[CONTEXT_KEY]
        if not context:
            return False
        return context and context.state.group == self.aiogd_intent_state_group


class IntentMiddlewareFactory:
    def __init__(self, storage: BaseStorage,
                 state_groups: Dict[str, Type[StatesGroup]]):
        super().__init__()
        self.storage = storage
        self.state_groups = state_groups

    def storage_proxy(self, data: dict):
        proxy = StorageProxy(
            bot=data['bot'],
            storage=self.storage,
            user_id=data['event_from_user'].id,
            chat_id=data['event_chat'].id,
            state_groups=self.state_groups,
        )
        return proxy

    async def process_message(
            self, handler: Callable, event: Message, data: dict,
    ):
        proxy = self.storage_proxy(data)
        stack = await proxy.load_stack()
        if stack.empty():
            context = None
        else:
            context = await proxy.load_context(stack.last_intent_id())
        data[STORAGE_KEY] = proxy
        data[STACK_KEY] = stack
        data[CONTEXT_KEY] = context
        return await handler(event, data)

    process_my_chat_member = process_message

    async def process_aiogd_update(
            self, handler: Callable, event: DialogUpdateEvent, data: dict,
    ):
        proxy = self.storage_proxy(data)
        data[STORAGE_KEY] = proxy
        if event.intent_id is not None:
            context = await proxy.load_context(event.intent_id)
            stack = await proxy.load_stack(context.stack_id)
        elif event.stack_id is not None:
            stack = await proxy.load_stack(event.stack_id)
            if stack.empty():
                if event.intent_id is not None:
                    raise OutdatedIntent(
                        stack.id,
                        f"Outdated intent id ({event.intent_id}) "
                        f"for stack ({stack.id})"
                    )
                context = None
            else:
                if event.intent_id is not None and event.intent_id != stack.last_intent_id():
                    raise OutdatedIntent(
                        stack.id,
                        f"Outdated intent id ({event.intent_id}) "
                        f"for stack ({stack.id})"
                    )
                context = await proxy.load_context(stack.last_intent_id())
        else:
            raise InvalidStackIdError(
                f"Both stack id and intent id are None: {event}"
            )

        data[STACK_KEY] = stack
        data[CONTEXT_KEY] = context
        return await handler(event, data)

    async def process_callback_query(
            self, handler: Callable, event: CallbackQuery, data: dict,
    ):
        proxy = self.storage_proxy(data)
        data[STORAGE_KEY] = proxy

        original_data = event.data
        intent_id, callback_data = remove_indent_id(event.data)
        if intent_id:
            context = await proxy.load_context(intent_id)
            stack = await proxy.load_stack(context.stack_id)
            if stack.last_intent_id() != intent_id:
                raise OutdatedIntent(
                    stack.id,
                    f"Outdated intent id ({intent_id}) for stack ({stack.id})"
                )
        else:
            stack = await proxy.load_stack()
            if stack.empty():
                context = None
            else:
                context = await proxy.load_context(stack.last_intent_id())
        data[STACK_KEY] = stack
        data[CONTEXT_KEY] = context
        data[CALLBACK_DATA_KEY] = original_data
        return await handler(event, data)


SUPPORTED_ERROR_EVENTS = {'message', 'callback_query', 'my_chat_member'}


async def context_saver_middleware(handler, event, data):
    result = await handler(event, data)
    proxy: StorageProxy = data.pop(STORAGE_KEY, None)
    if proxy:
        await proxy.save_context(data.pop(CONTEXT_KEY))
        await proxy.save_stack(data.pop(STACK_KEY))
    return result


class IntentErrorMiddleware(BaseMiddleware):
    def __init__(self, storage: BaseStorage,
                 state_groups: Dict[str, Type[StatesGroup]]):
        super().__init__()
        self.storage = storage
        self.state_groups = state_groups

    async def __call__(
            self,
            handler: Callable[
                [Union[Update, DialogUpdate], Dict[str, Any]], Awaitable[Any]],
            event: Update,
            data: Dict[str, Any],
    ) -> Any:

        try:
            error = data["exception"]
            if isinstance(error, InvalidStackIdError):
                return
            if event.event_type not in SUPPORTED_ERROR_EVENTS:
                return

            chat = data['event_chat']

            proxy = StorageProxy(
                bot=data['bot'],
                storage=self.storage,
                user_id=data['event_from_user'].id,
                chat_id=chat.id,
                state_groups=self.state_groups,
            )
            data[STORAGE_KEY] = proxy

            if isinstance(error, OutdatedIntent):
                stack = await proxy.load_stack(stack_id=error.stack_id)
            else:
                stack = await proxy.load_stack()
            if stack.empty():
                context = None
            else:
                context = await proxy.load_context(stack.last_intent_id())
            data[STACK_KEY] = stack
            data[CONTEXT_KEY] = context

            return await handler(event, data)
        finally:
            proxy: StorageProxy = data.pop(STORAGE_KEY, None)
            if not proxy:
                return
            context = data.pop(CONTEXT_KEY)
            if context is not None:
                await proxy.save_context(context)
            await proxy.save_stack(data.pop(STACK_KEY))



from typing import Optional, Any

from aiogram.dispatcher.filters.state import State

from aiogram_dialog.deprecation_utils import manager_deprecated
from aiogram_dialog.manager.protocols import (
    ManagedDialogProto, DialogManager, ManagedDialogAdapterProto,
)


class ManagedDialogAdapter(ManagedDialogAdapterProto):
    def __init__(self, dialog: ManagedDialogProto, manager: DialogManager):
        self.dialog = dialog
        self.manager = manager

    async def show(self, preview: bool = False,
                   manager: Optional[DialogManager] = None):
        manager_deprecated(manager)
        return await self.dialog.show(self.manager)

    async def next(self,
                   manager: Optional[DialogManager] = None):
        manager_deprecated(manager)
        return await self.dialog.next(self.manager)

    async def back(self,
                   manager: Optional[DialogManager] = None):
        manager_deprecated(manager)
        return await self.dialog.back(self.manager)

    async def switch_to(self, state: State,
                        manager: Optional[DialogManager] = None):
        manager_deprecated(manager)
        return await self.dialog.switch_to(state, self.manager)

    def find(self, widget_id,
             manager: Optional[DialogManager] = None) -> Optional[Any]:
        manager_deprecated(manager)
        widget = self.dialog.find(widget_id)
        if widget is None:
            return None
        return widget.managed(self.manager)



from copy import copy
from typing import Dict, Type, Optional

from aiogram import Bot
from aiogram.dispatcher.fsm.state import State, StatesGroup
from aiogram.dispatcher.fsm.storage.base import BaseStorage, StorageKey

from .context import Context
from .stack import Stack, DEFAULT_STACK_ID
from ..exceptions import UnknownState, UnknownIntent


class StorageProxy:
    def __init__(self, storage: BaseStorage,
                 user_id: int, chat_id: int, bot: Bot,
                 state_groups: Dict[str, Type[StatesGroup]]):
        self.storage = storage
        self.state_groups = state_groups
        self.user_id = user_id
        self.chat_id = chat_id
        self.bot = bot

    async def load_context(self, intent_id: str) -> Context:
        data = await self.storage.get_data(
            bot=self.bot,
            key=self._context_key(intent_id),
        )
        if not data:
            raise UnknownIntent(
                f"Context not found for intent id: {intent_id}")
        data["state"] = self._state(data["state"])
        return Context(**data)

    async def load_stack(self, stack_id: str = DEFAULT_STACK_ID) -> Stack:
        data = await self.storage.get_data(
            bot=self.bot,
            key=self._stack_key(stack_id),
        )
        if not data:
            return Stack(_id=stack_id)
        return Stack(**data)

    async def save_context(self, context: Optional[Context]) -> None:
        if not context:
            return
        data = copy(vars(context))
        data["state"] = data["state"].state
        await self.storage.set_data(
            bot=self.bot,
            key=self._context_key(context.id),
            data=data,
        )

    async def remove_context(self, intent_id: str):
        await self.storage.set_data(
            bot=self.bot,
            key=self._context_key(intent_id),
            data=dict(),
        )

    async def remove_stack(self, stack_id: str):
        await self.storage.set_data(
            bot=self.bot,
            key=self._stack_key(stack_id),
            data=dict(),
        )

    async def save_stack(self, stack: Optional[Stack]) -> None:
        if not stack:
            return
        if stack.empty() and not stack.last_message_id:
            await self.storage.set_data(
                bot=self.bot,
                key=self._stack_key(stack.id),
                data=dict(),
            )
        else:
            data = copy(vars(stack))
            await self.storage.set_data(
                bot=self.bot,
                key=self._stack_key(stack.id),
                data=data,
            )

    def _context_key(self, intent_id: str) -> StorageKey:
        return StorageKey(
            bot_id=self.bot.id,
            chat_id=self.chat_id,
            user_id=self.user_id,
            destiny=f"aiogd:context:{intent_id}",
        )

    def _stack_key(self, stack_id: str) -> StorageKey:
        return StorageKey(
            bot_id=self.bot.id,
            chat_id=self.chat_id,
            user_id=self.user_id,
            destiny=f"aiogd:stack:{stack_id}",
        )

    def _state(self, state: str) -> State:
        group, *_ = state.partition(":")
        for real_state in self.state_groups[group].__all_states__:
            if real_state.state == state:
                return real_state
        raise UnknownState(f"Unknown state {state}")



from abc import ABC
from calendar import monthcalendar
from datetime import date
from time import mktime
from typing import List, Callable, Union, Awaitable, TypedDict, Optional

from aiogram.types import InlineKeyboardButton, CallbackQuery

from aiogram_dialog.context.events import ChatEvent
from aiogram_dialog.dialog import Dialog
from aiogram_dialog.manager.protocols import DialogManager
from aiogram_dialog.widgets.widget_event import WidgetEventProcessor, ensure_event_processor
from .base import Keyboard
from ..managed import ManagedWidgetAdapter
from ...deprecation_utils import manager_deprecated

OnDateSelected = Callable[[ChatEvent, "ManagedCalendarAdapter", DialogManager, date], Awaitable]

# Constants for managing widget rendering scope
SCOPE_DAYS = "SCOPE_DAYS"
SCOPE_MONTHS = "SCOPE_MONTHS"
SCOPE_YEARS = "SCOPE_YEARS"

# Constants for scrolling months
MONTH_NEXT = "+"
MONTH_PREV = "-"

# Constants for prefixing month and year values
PREFIX_MONTH = "MONTH"
PREFIX_YEAR = "YEAR"

MONTHS_NUMBERS = [(1, 2, 3), (4, 5, 6), (7, 8, 9), (10, 11, 12)]


class CalendarData(TypedDict):
    current_scope: str
    current_offset: str


class Calendar(Keyboard, ABC):
    def __init__(self,
                 id: str,
                 on_click: Union[OnDateSelected, WidgetEventProcessor, None] = None,
                 when: Union[str, Callable] = None):
        super().__init__(id, when)
        self.on_click = ensure_event_processor(on_click)

    async def render_keyboard(self,
                              data,
                              manager: DialogManager) -> List[List[InlineKeyboardButton]]:
        offset = self.get_offset(manager)
        current_scope = self.get_scope(manager)

        if current_scope == SCOPE_DAYS:
            return self.days_kbd(offset)
        elif current_scope == SCOPE_MONTHS:
            return self.months_kbd(offset)
        elif current_scope == SCOPE_YEARS:
            return self.years_kbd(offset)

    async def process_callback(self,
                               c: CallbackQuery,
                               dialog: Dialog,
                               manager: DialogManager) -> bool:
        prefix = f"{self.widget_id}:"
        if not c.data.startswith(prefix):
            return False
        current_offset = self.get_offset(manager)
        data = c.data[len(prefix):]

        if data == MONTH_NEXT:
            new_offset = date(
                year=current_offset.year + (current_offset.month // 12),
                month=((current_offset.month % 12) + 1),
                day=1,
            )
            self.set_offset(new_offset, manager)

        elif data == MONTH_PREV:
            if current_offset.month == 1:
                new_offset = date(current_offset.year - 1, 12, 1)
                self.set_offset(new_offset, manager)
            else:
                new_offset = date(current_offset.year, (current_offset.month - 1), 1)
                self.set_offset(new_offset, manager)

        elif data in [SCOPE_MONTHS, SCOPE_YEARS]:
            self.set_scope(data, manager)

        elif data.startswith(PREFIX_MONTH):
            data = int(c.data[len(prefix) + len(PREFIX_MONTH):])
            new_offset = date(current_offset.year, data, 1)
            self.set_scope(SCOPE_DAYS, manager)
            self.set_offset(new_offset, manager)

        elif data.startswith(PREFIX_YEAR):
            data = int(c.data[len(prefix) + len(PREFIX_YEAR):])
            new_offset = date(data, 1, 1)
            self.set_scope(SCOPE_MONTHS, manager)
            self.set_offset(new_offset, manager)

        else:
            raw_date = int(data)
            await self.on_click.process_event(
                c, self.managed(manager), manager,
                date.fromtimestamp(raw_date),
            )
        return True

    def years_kbd(self, offset) -> List[List[InlineKeyboardButton]]:
        years = []
        for n in range(offset.year - 7, offset.year + 7, 3):
            year_row = []
            for year in range(n, n + 3):
                year_row.append(InlineKeyboardButton(text=str(year),
                                                     callback_data=f"{self.widget_id}:{PREFIX_YEAR}{year}"))
            years.append(year_row)
        return years

    def months_kbd(self, offset) -> List[List[InlineKeyboardButton]]:
        header_year = offset.strftime("year %Y")
        months = []
        for n in MONTHS_NUMBERS:
            season = []
            for month in n:
                month_text = date(offset.year, month, 1).strftime("%B")
                season.append(InlineKeyboardButton(text=month_text,
                                                   callback_data=f"{self.widget_id}:{PREFIX_MONTH}{month}"))
            months.append(season)
        return [
            [
                InlineKeyboardButton(text=header_year,
                                     callback_data=f"{self.widget_id}:{SCOPE_YEARS}"),
            ],
            *months
        ]

    def days_kbd(self, offset) -> List[List[InlineKeyboardButton]]:
        header_week = offset.strftime("%B %Y")
        weekheader = [InlineKeyboardButton(text=dayname, callback_data=" ")
                      for dayname in ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]]
        days = []
        for week in monthcalendar(offset.year, offset.month):
            week_row = []
            for day in week:
                if day == 0:
                    week_row.append(InlineKeyboardButton(text=" ",
                                                         callback_data=" "))
                else:
                    raw_date = int(mktime(date(offset.year, offset.month, day).timetuple()))
                    week_row.append(InlineKeyboardButton(text=str(day),
                                                         callback_data=f"{self.widget_id}:{raw_date}"))
            days.append(week_row)
        return [
            [
                InlineKeyboardButton(text=header_week,
                                     callback_data=f"{self.widget_id}:{SCOPE_MONTHS}"),
            ],
            weekheader,
            *days,
            [
                InlineKeyboardButton(text="Prev month",
                                     callback_data=f"{self.widget_id}:{MONTH_PREV}"),
                InlineKeyboardButton(text="Zoom out",
                                     callback_data=f"{self.widget_id}:{SCOPE_MONTHS}"),
                InlineKeyboardButton(text="Next month",
                                     callback_data=f"{self.widget_id}:{MONTH_NEXT}"),
            ],
        ]

    def get_scope(self, manager: DialogManager) -> str:
        calendar_data: CalendarData = manager.current_context().widget_data.get(self.widget_id, {})
        current_scope = calendar_data.get("current_scope")
        return current_scope or SCOPE_DAYS

    def get_offset(self, manager: DialogManager) -> date:
        calendar_data: CalendarData = manager.current_context().widget_data.get(self.widget_id, {})
        current_offset = calendar_data.get("current_offset")
        if current_offset is None:
            return date.today()
        return date.fromisoformat(current_offset)

    def set_offset(self, new_offset: date, manager: DialogManager) -> None:
        data = manager.current_context().widget_data.setdefault(self.widget_id, {})
        data["current_offset"] = new_offset.isoformat()

    def set_scope(self, new_scope: str, manager: DialogManager) -> None:
        data = manager.current_context().widget_data.setdefault(self.widget_id, {})
        data["current_scope"] = new_scope

    def managed(self, manager: DialogManager):
        return ManagedCalendarAdapter(self, manager)


class ManagedCalendarAdapter(ManagedWidgetAdapter[Calendar]):
    def get_scope(self, manager: Optional[DialogManager] = None) -> str:
        manager_deprecated(manager)
        return self.widget.get_scope(self.manager)

    def get_offset(self, manager: Optional[DialogManager] = None) -> date:
        manager_deprecated(manager)
        return self.widget.get_offset(self.manager)

    def set_offset(self, new_offset: date,
                   manager: Optional[DialogManager] = None) -> None:
        manager_deprecated(manager)
        return self.widget.set_offset(new_offset, self.manager)

    def set_scope(self, new_scope: str,
                  manager: Optional[DialogManager] = None) -> None:
        manager_deprecated(manager)
        return self.widget.set_scope(new_scope, self.manager)



from typing import Callable, Awaitable, Union, Sequence

from aiogram.dispatcher.filters.content_types import ContentTypesFilter
from aiogram.types import Message, ContentType

from aiogram_dialog.dialog import Dialog
from aiogram_dialog.manager.protocols import (
    DialogManager, ManagedDialogAdapterProto,
)
from aiogram_dialog.widgets.action import Actionable
from aiogram_dialog.widgets.widget_event import (
    WidgetEventProcessor, ensure_event_processor,
)

MessageHandlerFunc = Callable[
    [Message, ManagedDialogAdapterProto, DialogManager],
    Awaitable,
]


class BaseInput(Actionable):
    async def process_message(self, m: Message, dialog: Dialog, manager: DialogManager) -> bool:
        raise NotImplementedError


class MessageInput(BaseInput):
    def __init__(self, func: Union[MessageHandlerFunc, WidgetEventProcessor, None],
                 content_types: Union[Sequence[str], str] = ContentType.ANY):
        super().__init__()
        self.func = ensure_event_processor(func)
        self.filter = ContentTypesFilter(content_types=content_types)

    async def process_message(self, message: Message, dialog: Dialog,
                              manager: DialogManager) -> bool:
        if not await self.filter(message):
            return False
        await self.func.process_event(message, manager.dialog(), manager)
        return True



from typing import List, Dict, Optional, Callable, Awaitable, Union

from aiogram.types import InlineKeyboardButton, CallbackQuery

from aiogram_dialog.deprecation_utils import manager_deprecated
from aiogram_dialog.dialog import Dialog, ChatEvent
from aiogram_dialog.manager.protocols import DialogManager
from aiogram_dialog.widgets.widget_event import WidgetEventProcessor, ensure_event_processor
from .base import Keyboard
from .group import Group
from ..managed import ManagedWidgetAdapter
from ..when import WhenCondition

OnStateChanged = Callable[
    [ChatEvent, "ManagedScrollingGroupAdapter", DialogManager],
    Awaitable,
]


class ScrollingGroup(Group):
    def __init__(self, *buttons: Keyboard, id: str, width: Optional[int] = None,
                 height: int = 0, when: WhenCondition = None,
                 on_page_changed: Union[OnStateChanged, WidgetEventProcessor, None] = None):
        super().__init__(*buttons, id=id, width=width, when=when)
        self.height = height
        self.on_page_changed = ensure_event_processor(on_page_changed)

    async def _render_keyboard(
            self, data: Dict, manager: DialogManager
    ) -> List[List[InlineKeyboardButton]]:
        kbd = await super()._render_keyboard(data, manager)
        pages = len(kbd) // self.height + bool(len(kbd) % self.height)
        last_page = pages - 1
        if pages == 0:
            return kbd
        current_page = min(last_page, self.get_page(manager))
        next_page = min(last_page, current_page + 1)
        prev_page = max(0, current_page - 1)
        pager = [[
            InlineKeyboardButton(text="1", callback_data=f"{self.widget_id}:0"),
            InlineKeyboardButton(text="<", callback_data=f"{self.widget_id}:{prev_page}"),
            InlineKeyboardButton(text=str(current_page + 1),
                                 callback_data=f"{self.widget_id}:{current_page}"),
            InlineKeyboardButton(text=">", callback_data=f"{self.widget_id}:{next_page}"),
            InlineKeyboardButton(text=str(last_page + 1),
                                 callback_data=f"{self.widget_id}:{last_page}"),
        ]]
        return kbd[current_page * self.height: (current_page + 1) * self.height] + pager

    async def process_callback(self, c: CallbackQuery, dialog: Dialog,
                               manager: DialogManager) -> bool:
        prefix = f"{self.widget_id}:"
        if not c.data.startswith(prefix):
            return await super().process_callback(c, dialog, manager)
        new_page = int(c.data[len(prefix):])
        await self.set_page(c, new_page, manager)
        return True

    def get_page(self, manager: DialogManager) -> int:
        return manager.current_context().widget_data.get(self.widget_id, 0)

    async def set_page(self, event: ChatEvent, page: int,
                       manager: DialogManager) -> None:
        manager.current_context().widget_data[self.widget_id] = page
        await self.on_page_changed.process_event(
            event, self.managed(manager), manager,
        )

    def managed(self, manager: DialogManager):
        return ManagedScrollingGroupAdapter(self, manager)


class ManagedScrollingGroupAdapter(ManagedWidgetAdapter[ScrollingGroup]):
    def get_page(self, manager: Optional[DialogManager] = None) -> int:
        manager_deprecated(manager)
        return self.widget.get_page(self.manager)

    async def set_page(self, event: ChatEvent, page: int,
                       manager: Optional[DialogManager] = None) -> None:
        manager_deprecated(manager)
        return await self.widget.set_page(
            event, page, self.manager
        )



from typing import Optional, Any, Dict

from aiogram.types import ContentType

from aiogram_dialog.manager.manager import DialogManager
from aiogram_dialog.utils import MediaAttachment
from .base import Media
from ..when import WhenCondition


class StaticMedia(Media):
    def __init__(
            self,
            *,
            path: Optional[str] = None,
            url: Optional[str] = None,
            type: ContentType = ContentType.PHOTO,
            media_params: Dict = None,
            when: WhenCondition = None,
    ):
        super().__init__(when)
        if not (url or path):
            raise ValueError("Neither url nor path are provided")
        self.type = type
        self.path = path
        self.url = url
        self.media_params = media_params or {}

    async def _render_media(
            self,
            data: Any,
            manager: DialogManager
    ) -> Optional[MediaAttachment]:
        return MediaAttachment(
            type=self.type,
            url=self.url,
            path=self.path,
            **self.media_params,
        )



from typing import TypeVar, Generic

from ..manager.protocols import (
    ManagedDialogProto, DialogManager, ManagedWidgetProto,
)


class ManagedWidget(ManagedWidgetProto):
    def managed(self, manager: DialogManager):
        return ManagedWidgetAdapter(self, manager)


W = TypeVar("W", bound=ManagedWidget)


class ManagedWidgetAdapter(ManagedWidgetProto, Generic[W]):
    def __init__(self, widget: W, manager: DialogManager):
        self.widget = widget
        self.manager = manager

    def __getattr__(self, item):
        return getattr(self.widget, item)



from datetime import datetime
from logging import getLogger
from typing import Any, Optional, Dict

from aiogram.dispatcher.fsm.state import State
from aiogram.types import User, Chat, Message, CallbackQuery, Document

from .bg_manager import BgManager
from .dialog import ManagedDialogAdapter
from .protocols import (
    DialogManager, BaseDialogManager, ShowMode, LaunchMode,
    ManagedDialogAdapterProto, ManagedDialogProto, DialogRegistryProto,
    NewMessage,
)
from ..context.context import Context
from ..context.events import ChatEvent, StartMode, Data, FakeChat, FakeUser
from ..context.intent_filter import CONTEXT_KEY, STORAGE_KEY, STACK_KEY
from ..context.stack import Stack, DEFAULT_STACK_ID
from ..context.storage import StorageProxy
from ..exceptions import IncorrectBackgroundError
from ..utils import remove_kbd, show_message

logger = getLogger(__name__)


class ManagerImpl(DialogManager):
    def __init__(self, event: ChatEvent, registry: DialogRegistryProto,
                 data: Dict):
        self.disabled = False
        self._registry = registry
        self.event = event
        self.data = data
        self.show_mode: ShowMode = ShowMode.AUTO

    @property
    def registry(self) -> DialogRegistryProto:
        return self._registry

    def check_disabled(self):
        if self.disabled:
            raise IncorrectBackgroundError(
                "Detected background access to dialog manager. "
                "Please use background manager available via `manager.bg()` "
                "method to access methods from background tasks"
            )

    def is_preview(self) -> bool:
        return False

    def dialog(self) -> ManagedDialogAdapterProto:
        return ManagedDialogAdapter(self._dialog(), self)

    def _dialog(self) -> ManagedDialogProto:
        self.check_disabled()
        current = self.current_context()
        if not current:
            raise RuntimeError
        return self.registry.find_dialog(current.state)

    def current_context(self) -> Optional[Context]:
        return self.data[CONTEXT_KEY]

    def current_stack(self) -> Optional[Stack]:
        return self.data[STACK_KEY]

    def storage(self) -> StorageProxy:
        return self.data[STORAGE_KEY]

    async def _remove_kbd(self) -> None:
        chat = self.data['event_chat']
        bot = self.data['bot']

        message = None
        if self.current_stack().last_message_id:
            message = Message(chat=chat,
                              message_id=self.current_stack().last_message_id,
                              date=datetime.now()  # ToDo: check this
                              )
        await remove_kbd(bot, message)
        self.current_stack().last_message_id = None

    async def done(self, result: Any = None) -> None:
        await self._dialog().process_close(result, self)
        old_context = self.current_context()
        await self.mark_closed()
        context = self.current_context()
        if not context:
            await self._remove_kbd()
            return
        dialog = self._dialog()
        await dialog.process_result(old_context.start_data, result, self)
        if context.id == self.current_context().id:
            await self._dialog().show(self)

    async def mark_closed(self) -> None:
        self.check_disabled()
        storage = self.storage()
        stack = self.current_stack()
        await storage.remove_context(stack.pop())
        if stack.empty():
            self.data[CONTEXT_KEY] = None
        else:
            intent_id = stack.last_intent_id()
            self.data[CONTEXT_KEY] = await storage.load_context(intent_id)

    async def start(
            self,
            state: State,
            data: Data = None,
            mode: StartMode = StartMode.NORMAL,
    ) -> None:
        self.check_disabled()
        if mode is StartMode.NORMAL:
            await self._start_normal(state, data)
        elif mode is StartMode.RESET_STACK:
            await self.reset_stack()
            await self._start_normal(state, data)
        elif mode is StartMode.NEW_STACK:
            await self._start_new_stack(state, data)
        else:
            raise ValueError(f"Unknown start mode: {mode}")

    async def reset_stack(self, remove_keyboard: bool = True) -> None:
        storage = self.storage()
        stack = self.current_stack()
        while not stack.empty():
            await storage.remove_context(stack.pop())
        if remove_keyboard:
            await self._remove_kbd()
        self.data[CONTEXT_KEY] = None

    async def _start_new_stack(self, state: State, data: Data = None) -> None:
        stack = Stack()
        await self.bg(stack_id=stack.id).start(state, data, StartMode.NORMAL)

    async def _start_normal(self, state: State, data: Data = None) -> None:
        stack = self.current_stack()
        old_dialog: Optional[ManagedDialogProto] = None
        if not stack.empty():
            old_dialog = self._dialog()
            if old_dialog.launch_mode is LaunchMode.EXCLUSIVE:
                raise ValueError(
                    "Cannot start dialog on top of one with launch_mode==SINGLE")

        new_dialog = self.registry.find_dialog(state)
        launch_mode = new_dialog.launch_mode
        if launch_mode in (LaunchMode.EXCLUSIVE, LaunchMode.ROOT):
            await self.reset_stack(remove_keyboard=False)
        if launch_mode is LaunchMode.SINGLE_TOP:
            if new_dialog is old_dialog:
                await self.storage().remove_context(stack.pop())

        await self.storage().save_context(self.current_context())
        context = stack.push(state, data)
        self.data[CONTEXT_KEY] = context
        await self._dialog().process_start(self, data, state)
        if context.id == self.current_context().id:
            await self._dialog().show(self)

    async def switch_to(self, state: State) -> None:
        self.check_disabled()
        context = self.current_context()
        if context.state.group != state.group:
            raise ValueError(f"Cannot switch to another state group. "
                             f"Current state: {context.state}, asked for {state}")
        context.state = state

    async def show(self, new_message: NewMessage) -> Message:
        stack = self.current_stack()
        bot = self.data['bot']
        chat = self.data['event_chat']
        if (
                isinstance(self.event, CallbackQuery)
                and self.event.message
                and stack.last_message_id == self.event.message.message_id
        ):
            old_message = self.event.message
        else:
            if stack and stack.last_message_id:
                if stack.last_media_id:
                    # we create document because
                    # * there is no method to set content type explicitly
                    # * we don't really care fo exact content type
                    document = Document(
                        file_id=stack.last_media_id,
                        file_unique_id=stack.last_media_unique_id,
                    )
                    text = None
                else:
                    document = None
                    # we set some non empty-text which is not equal to anything
                    text = "𝔞𝔦𝔬𝔤𝔯𝔞𝔪 𝔡𝔦𝔞𝔩𝔬𝔤 𝔲𝔫𝔦𝔮𝔲𝔢 𝔱𝔢𝔵𝔱"
                old_message = Message(message_id=stack.last_message_id,
                                      document=document,
                                      text=text,
                                      chat=chat,
                                      date=datetime.now())
            else:
                old_message = None
        if new_message.show_mode is ShowMode.AUTO:
            new_message.show_mode = self._calc_show_mode()
        res = await show_message(bot, new_message, old_message)
        if isinstance(self.event, Message):
            stack.last_income_media_group_id = self.event.media_group_id
        self.show_mode = ShowMode.EDIT
        return res

    def _calc_show_mode(self) -> ShowMode:
        if self.show_mode is not ShowMode.AUTO:
            return self.show_mode
        if isinstance(self.event, Message):
            if self.event.media_group_id is None:
                return ShowMode.SEND
            elif self.event.media_group_id == self.current_stack().last_income_media_group_id:
                return ShowMode.EDIT
            else:
                return ShowMode.SEND
        return ShowMode.EDIT

    async def update(self, data: Dict) -> None:
        self.current_context().dialog_data.update(data)
        await self._dialog().show(self)

    def bg(
            self,
            user_id: Optional[int] = None,
            chat_id: Optional[int] = None,
            stack_id: Optional[str] = None,
            load: bool = False,
    ) -> "BaseDialogManager":
        current_chat = self.data['event_chat']
        current_user = self.event.from_user
        if chat_id in (None, current_chat.id):
            chat = current_chat
        else:
            chat = FakeChat(id=chat_id, type="")

        if user_id in (None, current_user.id):
            user = current_user
        else:
            user = FakeUser(id=user_id, is_bot=False, first_name="")

        same_chat = (user.id == current_user.id and chat.id == current_chat.id)
        intent_id = None
        if stack_id is None:
            if same_chat:
                stack_id = self.current_stack().id
                if self.current_context():
                    intent_id = self.current_context().id
            else:
                stack_id = DEFAULT_STACK_ID

        return BgManager(
            user=user,
            chat=chat,
            bot=self.data['bot'],
            registry=self.registry,
            intent_id=intent_id,
            stack_id=stack_id,
            load=load,
        )

    async def close_manager(self) -> None:
        self.check_disabled()
        self.disabled = True
        del self._registry
        del self.event
        del self.data



from typing import Optional, Any

from aiogram_dialog.manager.manager import DialogManager
from aiogram_dialog.utils import MediaAttachment
from aiogram_dialog.widgets.when import Whenable


class Media(Whenable):
    async def render_media(
            self,
            data: Any,
            manager: DialogManager
    ) -> Optional[MediaAttachment]:
        if not self.is_(data, manager):
            return None
        return await self._render_media(data, manager)

    async def _render_media(
            self,
            data: Any,
            manager: DialogManager
    ) -> Optional[MediaAttachment]:
        return None



from dataclasses import dataclass
from enum import Enum
from typing import Optional, Any, Protocol, Union, Type, Dict

from aiogram import Bot, Router
from aiogram.dispatcher.fsm.state import State, StatesGroup
from aiogram.types import (
    ContentType, Message, Chat, InlineKeyboardMarkup,
)

from ..context.context import Context
from ..context.events import StartMode, ChatEvent, Data, DialogUpdate
from ..context.stack import Stack


class ShowMode(Enum):
    AUTO = "auto"
    EDIT = "edit"
    SEND = "send"


class LaunchMode(Enum):
    """
    `ROOT` dialogs will be always a root dialog in stack.
        Starting such dialogs will automatically reset stack.
        Example: main menu

    `EXCLUSIVE` dialogs can be only a single dialog in stack.
        Starting such dialogs will automatically reset stack.
        Starting other dialogs on top of them is forbidden
        Example: Banners

    `SINGLE_TOP` dialogs will not be repeated on top of stack.
        Starting the same dialog right on top of it will just replace it.
        Example: product page

    `STANDARD` dialogs have no limitations themselves
    """
    STANDARD = "standard"
    ROOT = "root"
    EXCLUSIVE = "exclusive"
    SINGLE_TOP = "single_top"


class ManagedDialogAdapterProto:
    async def show(self):
        pass

    async def next(self):
        pass

    async def back(self):
        pass

    async def switch_to(self, state: State):
        pass

    def find(self, widget_id) -> Optional[Any]:
        pass


class ManagedDialogProto(Protocol):
    launch_mode: LaunchMode

    def register(self, registry: "DialogRegistryProto", router: Router, *args,
                 **kwargs) -> None:
        pass

    def states_group_name(self) -> str:
        pass

    def states_group(self) -> Type[StatesGroup]:
        pass

    async def process_close(self, result: Any, manager: "DialogManager"):
        pass

    async def process_start(self, manager: "DialogManager", start_data: Any,
                            state: Optional[State] = None) -> None:
        pass

    async def show(self, manager: "DialogManager"):
        pass

    async def process_result(self, start_data: Data, result: Any,
                             manager: "DialogManager"):
        pass

    async def next(self, manager: "DialogManager"):
        pass

    async def back(self, manager: "DialogManager"):
        pass

    async def switch_to(self, state: State, manager: "DialogManager"):
        pass

    def find(self, widget_id) -> Optional["ManagedWidgetProto"]:
        pass


@dataclass
class MediaId:
    file_id: str
    file_unique_id: Optional[str] = None

    def __eq__(self, other):
        if type(other) is not MediaId:
            return False
        if self.file_unique_id is None or other.file_unique_id is None:
            return self.file_id == other.file_id
        return self.file_unique_id == other.file_unique_id


class MediaIdStorageProtocol(Protocol):
    async def get_media_id(
            self,
            path: Optional[str],
            url: Optional[str],
            type: ContentType,
    ) -> Optional[MediaId]:
        raise NotImplementedError

    async def save_media_id(
            self,
            path: Optional[str],
            url: Optional[str],
            type: ContentType,
            media_id: MediaId,
    ) -> None:
        raise NotImplementedError


class MediaAttachment:
    def __init__(
            self,
            type: ContentType,
            url: Optional[str] = None,
            path: Optional[str] = None,
            file_id: Optional[MediaId] = None,
            **kwargs,
    ):
        if not (url or path or file_id):
            raise ValueError("Neither url nor path not file_id are provided")
        self.type = type
        self.url = url
        self.path = path
        self.file_id = file_id
        self.kwargs = kwargs


@dataclass
class NewMessage:
    chat: Chat
    text: Optional[str] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    parse_mode: Optional[str] = None
    show_mode: ShowMode = ShowMode.AUTO
    disable_web_page_preview: Optional[bool] = None
    media: Optional[MediaAttachment] = None


class DialogRegistryProto(Protocol):
    def find_dialog(self, state: Union[State, str]) -> ManagedDialogProto:
        pass

    async def notify(self, bot: Bot, update: DialogUpdate) -> None:
        pass

    @property
    def media_id_storage(self) -> MediaIdStorageProtocol:
        raise NotImplementedError


class BaseDialogManager(Protocol):
    event: ChatEvent

    @property
    def registry(self) -> DialogRegistryProto:
        raise NotImplementedError

    async def done(self, result: Any = None) -> None:
        pass

    async def mark_closed(self) -> None:
        pass

    async def start(
            self,
            state: State,
            data: Data = None,
            mode: StartMode = StartMode.NORMAL,
    ) -> None:
        pass

    async def switch_to(self, state: State) -> None:
        pass

    async def update(self, data: Dict) -> None:
        pass

    def bg(
            self,
            user_id: Optional[int] = None,
            chat_id: Optional[int] = None,
            stack_id: Optional[str] = None,
            load: bool = False,  # load chat and user
    ) -> "BaseDialogManager":
        pass


class DialogManager(BaseDialogManager):
    event: ChatEvent  # current processing event
    data: Dict  # data from middleware
    show_mode: ShowMode  # mode used to show messages

    def is_preview(self) -> bool:
        pass

    def current_context(self) -> Optional[Context]:
        pass

    def current_stack(self) -> Optional[Stack]:
        pass

    def dialog(self) -> ManagedDialogAdapterProto:
        pass

    async def close_manager(self) -> None:
        pass

    async def show(self, new_message: NewMessage) -> Message:
        pass

    async def reset_stack(self, remove_keyboard: bool = True) -> None:
        pass


class ManagedWidgetProto(Protocol):
    def managed(self, manager: DialogManager) -> Any:
        pass



import asyncio
from contextvars import copy_context
from typing import Sequence, Type, Dict, Optional

from aiogram import Dispatcher, Bot, Router
from aiogram.dispatcher.event.telegram import TelegramEventObserver
from aiogram.dispatcher.filters import Command
from aiogram.dispatcher.fsm.state import State, StatesGroup, any_state
from aiogram.types import User, Chat, Message

from .manager_middleware import ManagerMiddleware
from .protocols import (
    ManagedDialogProto, DialogRegistryProto, DialogManager,
    MediaIdStorageProtocol,
)
from .update_handler import handle_update
from ..context.events import StartMode, DIALOG_EVENT_NAME, DialogUpdate
from ..context.intent_filter import (
    IntentFilter, IntentMiddlewareFactory, IntentErrorMiddleware,
    context_saver_middleware,
)
from ..context.media_storage import MediaIdStorage


class DialogEventObserver(TelegramEventObserver):
    pass


class DialogRegistry(DialogRegistryProto):
    def __init__(
            self,
            dp: Dispatcher,
            dialogs: Sequence[ManagedDialogProto] = (),
            media_id_storage: Optional[MediaIdStorageProtocol] = None,
    ):
        self.dp = dp
        self.update_handler = self.dp.observers[DIALOG_EVENT_NAME] = DialogEventObserver(
            router=self.dp, event_name=DIALOG_EVENT_NAME
        )

        self.dialogs = {
            d.states_group(): d for d in dialogs
        }
        self.state_groups: Dict[str, Type[StatesGroup]] = {
            d.states_group_name(): d.states_group() for d in dialogs
        }
        self.register_update_handler(handle_update, any_state)

        observer: TelegramEventObserver
        for observer in self.dp.observers.values():
            observer.bind_filter(IntentFilter)

        self._register_middleware()
        if media_id_storage is None:
            media_id_storage = MediaIdStorage()
        self._media_id_storage = media_id_storage

    @property
    def media_id_storage(self) -> MediaIdStorageProtocol:
        return self._media_id_storage

    def register(self, dialog: ManagedDialogProto, *args, router: Router = None, **kwargs):
        group = dialog.states_group()
        if group in self.dialogs:
            raise ValueError(f"StatesGroup `{group}` is already used")
        self.dialogs[group] = dialog
        self.state_groups[dialog.states_group_name()] = group
        dialog.register(
            self,
            router if router else self.dp,
            *args,
            aiogd_intent_state_group=group,
            **kwargs
        )

    def register_start_handler(self, state: State):
        async def start_dialog(m: Message, dialog_manager: DialogManager):
            await dialog_manager.start(state, mode=StartMode.RESET_STACK)

        self.dp.message.register(start_dialog, Command(commands="start"), any_state)

    def _register_middleware(self):
        manager_middleware = ManagerMiddleware(self)
        intent_middleware = IntentMiddlewareFactory(
            storage=self.dp.fsm.storage, state_groups=self.state_groups
        )
        self.dp.message.middleware(manager_middleware)
        self.dp.callback_query.middleware(manager_middleware)
        self.update_handler.middleware(manager_middleware)
        self.dp.my_chat_member.middleware(manager_middleware)
        self.dp.errors.middleware(manager_middleware)

        self.dp.message.outer_middleware(intent_middleware.process_message)
        self.dp.callback_query.outer_middleware(intent_middleware.process_callback_query)
        self.update_handler.outer_middleware(intent_middleware.process_aiogd_update)
        self.dp.my_chat_member.outer_middleware(intent_middleware.process_my_chat_member)

        self.dp.message.middleware(context_saver_middleware)
        self.dp.callback_query.middleware(context_saver_middleware)
        self.update_handler.middleware(context_saver_middleware)
        self.dp.my_chat_member.middleware(context_saver_middleware)

        self.dp.errors.outer_middleware(IntentErrorMiddleware(
            storage=self.dp.fsm.storage, state_groups=self.state_groups
        ))


    def find_dialog(self, state: State) -> ManagedDialogProto:
        return self.dialogs[state.group]

    def register_update_handler(self, callback, *custom_filters, **kwargs) -> None:
        self.update_handler.register(
            callback, *custom_filters, **kwargs
        )

    async def notify(self, bot: Bot, update: DialogUpdate) -> None:
        callback = lambda: asyncio.create_task(self._process_update(bot, update))

        asyncio.get_running_loop().call_soon(
            callback,
            context=copy_context()
        )

    async def _process_update(self, bot: Bot, update: DialogUpdate):
        event = update.event
        Bot.set_current(bot)
        User.set_current(event.from_user)
        Chat.set_current(event.chat)
        await self.dp.propagate_event(
            update_type="update",
            event=update,
            bot=bot,
            event_from_user=event.from_user,
            event_chat=event.chat,
        )



from logging import getLogger
from typing import Dict, Optional, List

from aiogram.dispatcher.fsm.state import State
from aiogram.types import InlineKeyboardMarkup, Message, CallbackQuery, UNSET

from .dialog import Dialog, DialogWindowProto
from .manager.protocols import DialogManager
from .utils import NewMessage, MediaAttachment
from .widgets.action import Actionable
from .widgets.data import PreviewAwareGetter
from .widgets.kbd import Keyboard
from .widgets.utils import (
    ensure_widgets, ensure_data_getter, GetterVariant, WidgetSrc,
)

logger = getLogger(__name__)


class Window(DialogWindowProto):
    def __init__(self,
                 *widgets: WidgetSrc,
                 state: State,
                 getter: GetterVariant = None,
                 parse_mode: Optional[str] = UNSET,
                 disable_web_page_preview: Optional[bool] = None,
                 preview_add_transitions: Optional[List[Keyboard]] = None,
                 preview_data: GetterVariant = None,
                 ):
        (
            self.text, self.keyboard, self.on_message, self.media,
        ) = ensure_widgets(widgets)
        self.getter = PreviewAwareGetter(
            ensure_data_getter(getter),
            ensure_data_getter(preview_data),
        )
        self.state = state
        self.parse_mode = parse_mode
        self.disable_web_page_preview = disable_web_page_preview
        self.preview_add_transitions = preview_add_transitions

    async def render_text(self, data: Dict, manager: DialogManager) -> str:
        return await self.text.render_text(data, manager)

    async def render_media(
            self, data: Dict,
            manager: DialogManager
    ) -> Optional[MediaAttachment]:
        if self.media:
            return await self.media.render_media(data, manager)

    async def render_kbd(self, data: Dict,
                         manager: DialogManager) -> InlineKeyboardMarkup:
        keyboard = await self.keyboard.render_keyboard(data, manager)
        return InlineKeyboardMarkup(inline_keyboard=keyboard)

    async def load_data(self, dialog: "Dialog",
                        manager: DialogManager) -> Dict:
        return await self.getter(**manager.data)

    async def process_message(self, message: Message, dialog: Dialog,
                              manager: DialogManager):
        if self.on_message:
            await self.on_message.process_message(message, dialog, manager)

    async def process_callback(self, c: CallbackQuery, dialog: Dialog,
                               manager: DialogManager):
        if self.keyboard:
            await self.keyboard.process_callback(c, dialog, manager)

    async def render(self, dialog: Dialog, manager: DialogManager) -> NewMessage:
        logger.debug("Show window: %s", self)
        chat = manager.data['event_chat']
        current_data = await self.load_data(dialog, manager)
        return NewMessage(
            chat=chat,
            text=await self.render_text(current_data, manager),
            reply_markup=await self.render_kbd(current_data, manager),
            parse_mode=self.parse_mode,
            disable_web_page_preview=self.disable_web_page_preview,
            media=await self.render_media(current_data, manager),
        )

    def get_state(self) -> State:
        return self.state

    def find(self, widget_id) -> Optional[Actionable]:
        if self.keyboard:
            res = self.keyboard.find(widget_id)
            if res:
                return res
        if self.on_message:
            return self.on_message.find(widget_id)
        return None

    def __repr__(self):
        return f"<{self.__class__.__qualname__}({self.state})>"
